<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Security;

use App\Entity\Actors\Actor;
use Psr\Log\LoggerInterface;
use App\Providers\ApiKeyProvider;
use App\Security\BaseAuthenticator;
use Doctrine\ORM\EntityManagerInterface;
use App\Providers\BasicFingerPrintProvider;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Passport;
use Symfony\Component\Security\Core\Exception\AuthenticationException;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Badge\UserBadge;
use Symfony\Component\Security\Http\Authenticator\Token\PostAuthenticationToken;
use Symfony\Component\Security\Http\Authenticator\Passport\SelfValidatingPassport;
use Symfony\Component\Security\Core\Exception\CustomUserMessageAuthenticationException;
use Lexik\Bundle\JWTAuthenticationBundle\TokenExtractor\AuthorizationHeaderTokenExtractor;

class BasicAuthenticator extends BaseAuthenticator
{

    protected $container;
    protected $userRepository;

    public function __construct(
        ContainerInterface $container,
        EntityManagerInterface $entityManager,
        protected TagAwareCacheInterface $cache,
        protected LoggerInterface $logger
    ) {
        $this->container = $container;
        $this->userRepository = $entityManager->getRepository(Actor::class);
    }

    /**
     * Called on every request to decide if this authenticator should be
     * used for the request. Returning `false` will cause this authenticator
     * to be skipped.
     */
    public function supports(Request $request): ?bool
    {
        return $request->headers->has('Authorization') && strpos($request->headers->get('Authorization'), 'Basic') === 0;
    }

    public function authenticate(Request $request): Passport
    {
        $tokenExtractor = new AuthorizationHeaderTokenExtractor('Basic', 'Authorization');

        $basicToken = $tokenExtractor->extract($request);
        if (null === $basicToken) {
            // The token header was empty, authentication fails with HTTP Status
            // Code 401 "Unauthorized"
            throw new CustomUserMessageAuthenticationException('No API token provided');
        }
        $basicProvider = new BasicFingerPrintProvider($this->container);
        $decodedToken = base64_decode($basicToken);
        [$username, $password] = explode(':', $decodedToken);
        $executeBasicAuth = true;
        if (str_starts_with($username, 'X-API-KEY-')) {
            try {
                $apiKeyProvider = new ApiKeyProvider($this->container, $this->cache, $this->logger);
                $userId = $apiKeyProvider->validateFingerprint($password, $request);
                $executeBasicAuth = false;
            } catch (AuthenticationException $ex) {
                # Proceed with validating credentials, as even username can have this as prefix
                $this->logger->debug("Attempting Basic Auth Validation");
            }
        }

        # Proceeding with Basic Auth validation of username and password default.
        if ($executeBasicAuth) {
            $decodedToken = $basicProvider->validateFingerprint($username, $password);
            $userId = $decodedToken['uuid'];
        }

        $passport = new SelfValidatingPassport(
            new UserBadge($userId, function ($userIdentifier) {
                return $this->userRepository->findOneBy(['userId' => $userIdentifier]);
            })
        );
        $passport->setAttribute('token', $basicToken);
        return $passport;
    }

    /**
     * Shortcut to create a PostAuthenticationToken for you, if you don't really
     * care about which authenticated token you're using.
     */
    public function createToken(Passport $passport, string $firewallName): TokenInterface
    {
        return new PostAuthenticationToken($passport->getUser(), $firewallName, $passport->getUser()->getRoles(true, true));
    }
}
