<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Exception;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Constants\ErrorConstants;
use App\Security\Voter\FsrVoter;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\Workflow\WorkflowCollection;
use App\Providers\PermissionsDictionaryBuilder;
use App\Service\ConfigExportImport\PlaybookHelperConfig;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class PreInstallConfig
 * @package App\Service\ConfigExportImport
 */
class PreInstallConfig extends PortingConfig
{

    public function __construct(
        private FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        protected LoggerInterface $logger,
        private PlaybookHelperConfig $playbookHelperConfig
    ) {
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * Loads data from tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @return array
     */
    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $data = $this->playbookHelperConfig->loadDataFromDirectory($key, $tmpDirectory);
        return $data;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * @param array $allData
     * @param string $key
     * @param string $tokenType
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];

        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')
        ) {
            return [];
        }

        $options = [];
        $preInstallCollection = array_key_exists('collections', $data) && is_array($data['collections']) && count($data['collections']) == 1 ? $data['collections'][0] : [];

        if ($preInstallCollection) {
            $collectionOption = $this->playbookHelperConfig->generatePlaybookCollectionOptions([$preInstallCollection], $this->importJob);
            $options['collections'] = [
                "include" => true,
                "values" => $collectionOption
            ];
        }

        return $options;
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return boolean
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        try {
            $preInstallCollection = array_key_exists('collections', $data) && is_array($data['collections']) && count($data['collections']) == 1 ? $data['collections'][0] : [];
            $executionList = array_key_exists('executionList', $data) ? $data['executionList'] : [];
            if ($preInstallCollection) {
                $collectionResult = $this->importPlaybookCollection($preInstallCollection, $tokenType, $importedBy);
                if ($collectionResult) {
                    $collectionRepository = $this->loadRepository(WorkflowCollection::class);
                    $collection = $collectionRepository->findOneBy(['name' => $preInstallCollection['name']]);
                    $collectionUuid = $collection->getUuid();
                    $preInstallWorkflows = $collection->getWorkflows();
                    $workflowUuidObjectMap = $this->playbookHelperConfig->mapWorkflowUuidObject($preInstallWorkflows);
                    foreach ($executionList as $playbookUuid) {
                        if (array_key_exists($playbookUuid, $workflowUuidObjectMap)) {
                            $taskId = $this->playbookHelperConfig->executePlaybook($workflowUuidObjectMap[$playbookUuid], $tokenType);
                            $status = 'awaiting';
                            while (!in_array($status, PlaybookHelperConfig::WORKFLOW_EXECUTION_STATUS)) {
                                $status = $this->playbookHelperConfig->checkWorkflowExecutionStatus($taskId);
                                if (in_array($status, PlaybookHelperConfig::WORKFLOW_EXECUTION_FAILED_STATUS)) {
                                    throw new Exception(sprintf(ErrorConstants::FSR_CH_0000023, "Pre-Install"));
                                } else {
                                    sleep(5);
                                }
                            }
                        }
                    }
                    $this->playbookHelperConfig->deleteCollection($collectionUuid, $tokenType);
                }
            }
        } catch (\Exception $exp) {
            $this->logger->error($exp->getMessage());
            throw new \Exception($exp->getMessage());
        }

        return true;
    }

    /**
     * Imports playbook collection
     *
     * @param array $collection
     * @return boolean
     */
    private function importPlaybookCollection(array $collection, $tokenType, $importedBy = [])
    {
        if ($collection['workflows']) {
            foreach ($collection['workflows'] as $playbook) {
                $playbook['importedBy'] = array_key_exists("importedBy", $playbook) && $playbook['importedBy'] ? array_merge($playbook['importedBy'], $importedBy) : $importedBy;
            }
        }

        $repository = $this->loadRepository(WorkflowCollection::class);
        $existingByUuid = $repository->findOneBy(['uuid' => $collection['uuid']]);

        if ($existingByUuid) {
            $this->playbookHelperConfig->deleteCollection($collection['uuid'], $tokenType);
        }

        $fullURI = sprintf('%s/api/3/workflow_collections', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
        $result = $this->sendRequest($fullURI, 'post', $collection, $tokenType);
        return $result;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        return [];
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        return [];
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        return [];
    }

    public function generateInfoContent(array $options)
    {
        return [];
    }
}
