<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use App\Entity\ModelMetadata\ModelMetadata;
use App\Entity\ViewTemplate\SystemViewTemplate;
use App\Service\UtilityService;
use Psr\Log\LoggerInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class ViewConfig
 * @package App\Service\ConfigExportImport
 */
class ViewConfig extends PortingConfig
{

    public function __construct(
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        protected LoggerInterface $logger,
        protected UtilityService $utilsService
    ) {
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }
    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $typeKey, $tokenType)
    {
        $data = $allData[$typeKey];
        $options = [];
        $repository = $this->loadRepository(SystemViewTemplate::class);
        $navigationItemReplaceOptions = [];
        if (array_key_exists('navigationItemReplaceOptions', $data)) {
            $navigationItemReplaceOptions = $data['navigationItemReplaceOptions'];
            unset($data['navigationItemReplaceOptions']);
        }
        $replaceNavigation = array_key_exists('replaceNavigation', $navigationItemReplaceOptions) &&
            $navigationItemReplaceOptions['replaceNavigation'] ?
            $navigationItemReplaceOptions['replaceNavigation'] : false;
        $navigationItemsToReplace = array_key_exists('navigationItemsToReplace', $navigationItemReplaceOptions) &&
            $navigationItemReplaceOptions['navigationItemsToReplace'] ?
            $navigationItemReplaceOptions['navigationItemsToReplace'] : [];
        foreach ($data as $key => $value) {
            if (!$value) {
                continue;
            }
            if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                // This is a sequential array, used by legacy module view templates
                foreach ($value as $view) {
                    $existing = $repository->findOneBy(['uuid' => $view['uuid']]);
                    $individualOptions = [];
                    $uuid = $view['uuid'];
                    if (!$uuid) {
                        $uuid = implode('-', explode('/', $view['id']));
                    }
                    $individualOptions['uuid'] = $uuid;
                    if ($view['config'] && array_key_exists('title', $view['config'])) {
                        $individualOptions['name'] = $view['config']['title'];
                    }
                    $individualOptions['module'] = $key;
                    $individualOptions['exists'] = $existing ? true : false;
                    $individualOptions['include'] = true;

                    $options[] = $individualOptions;
                }
                continue;
            }
            $existing = $repository->findOneBy(['uuid' => $value['uuid']]);
            $individualOptions = [];
            $uuid = $value['uuid'];
            $pieces = explode('-', $value['uuid']);
            if (count($pieces) === 3) {
                $individualOptions['module'] = $pieces[1];
            }
            if ($value['config'] && array_key_exists('title', $value['config'])) {
                $individualOptions['name'] = $value['config']['title'];
            }

            if ($value['uuid'] == 'app') {
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    foreach ($existing->getConfig()['navigation'] as $existingItem) {
                        if ($existingItem['title'] == $item['title']) {
                            $value['config']['navigation'][$key]['exists'] = true;
                            $value['config']['navigation'][$key]['include'] = true;
                            $itemMergeType = in_array($item['title'], $navigationItemsToReplace) ? AppConstants::MERGE_TYPES['replace'] : AppConstants::MERGE_TYPES['merge'];
                            $value['config']['navigation'][$key]['mergeType'] = $itemMergeType;
                        }
                    }
                }
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    if (!array_key_exists("exists", $value['config']['navigation'][$key]) || (array_key_exists("exists", $value['config']['navigation'][$key]) && $value['config']['navigation'][$key]['exists'] == false)) {
                        $value['config']['navigation'][$key]['exists'] = false;
                        $value['config']['navigation'][$key]['include'] = true;
                        $value['config']['navigation'][$key]['mergeType'] = AppConstants::MERGE_TYPES['merge'];
                    }
                }
                $replaceEntireNavigation = $replaceNavigation && count($navigationItemsToReplace) == 0 ? true : false;
                $individualOptions['replaceEntireNavigation'] = $replaceEntireNavigation;

                // If all navigation items are intended to replace then global mergeType will be replace
                // If all navigation items are intended to append then global mergeType will be append
                // If some navigation items are having mergeType as replace and some having append then global mergeType will be empty string
                $individualOptions['mergeType'] = $replaceNavigation ? (count($navigationItemsToReplace) == 0 ? AppConstants::MERGE_TYPES['replace'] : "") : AppConstants::MERGE_TYPES['merge'];
            }
            if (is_array($pieces) && array_key_exists("2", $pieces) && $pieces[2] == 'settings') {
                $individualOptions['_hidden'] = true;
            }
            $individualOptions['uuid'] = $uuid;
            $individualOptions['exists'] = $existing ? true : false;
            $individualOptions['include'] = true;
            $individualOptions['config'] = $value['config'];

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        foreach ($options as $option) {
            if (is_array($option) && array_key_exists("_hidden", $option) && $option['_hidden'] && strpos($option['uuid'], '-settings')) {
                $detailUuid = str_replace('settings', 'detail', $option['uuid']);
                $detailView = self::findBy($options, 'uuid', $detailUuid);
                $option['include'] = $detailView['include'];
            }
            if (!$option['include']) {
                // Do not merge skipped items
                continue;
            }
            # When imported through import wizard, 'mergeType' value selected in wizard is honoured
            $option['appendNavigation'] = array_key_exists('mergeType', $option) && ($option['mergeType'] == AppConstants::MERGE_TYPES['replace'] || $option['mergeType'] == '') ? false : true;

            $foundView = null;
            foreach ($data as $value) {
                if (!$value) {
                    continue;
                }
                if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                    // This is a sequential array, used by module view templates
                    foreach ($value as $view) {
                        if ($option['uuid'] == $view['uuid']) {
                            $foundView = $view;
                            break;
                        }
                    }
                    continue;
                }
                if (!array_key_exists('uuid', $value)) {
                    $value['uuid'] = null;
                }
                $importJobType = $this->importJob ? $this->importJob->getType() : AppConstants::IMPORT_JOB_TYPES["IMPORT_WIZARD"];
                $productType = $this->utilsService->getRequiredLicenseDetails(CacheConstants::PRODUCT_TYPE);
                if ("app" == $value['uuid']) {
                    if ($productType == AppConstants::PRODUCT_TYPE['fortitip'] && $importJobType == 'SolutionPack Import' && !$this->containsFortiTIP($importedBy))
                        continue;
                    $fullURI = sprintf('%s/api/views/1/app', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                    $oldNav = json_decode($this->sendRequest($fullURI, 'GET', null, $tokenType), true);
                    if ($option['appendNavigation']) {
                        $foundView = $this->mergeNAppendNav($option['config']['navigation'], $oldNav);
                    } else {
                        $foundView = $this->replaceNavItems($option, $oldNav);
                    }
                } elseif ($option['uuid'] == $value['uuid']) {
                    $value['importedBy'] = array_key_exists("importedBy", $value) && $value['importedBy'] ? array_merge($value['importedBy'], $importedBy) : $importedBy;
                    $foundView = $value;
                    break;
                }
            }
            if ($option['exists'] && $foundView) {
                $toUpdate[] = $foundView;
            } elseif ($foundView) {
                $toAdd[] = $foundView;
            }
        }
        $results = [];
        if (!empty($toAdd)) {
            $fullURI = sprintf('%s/api/3/', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $toAdd, $tokenType);
            $results[] = $result;
        }
        if (!empty($toUpdate)) {
            $fullURI = sprintf('%s/api/3/update/system_view_templates', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        return $this->getResourceFromIRI($options, SystemViewTemplate::class);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modifiedOptions = [];
        $appendNavigation = [];
        $navigationOptions = [];
        $navigationTitleMergeTypeMap = [];
        $metaData = [];
        $key = null;
        foreach ($options as $value) {
            if (is_string($value)) {
                $modifiedOptions[] = $value;
            }
            if (is_array($value) && !empty($value)) {
                $modifiedOptions[] = $value['value'];
                if ($value['value'] == "app") {
                    $appendNavigation = array_key_exists('appendNavigation', $value) && !empty($value['appendNavigation']) ? $value['appendNavigation'] : [];
                    $navigationOptions = array_key_exists('navigationOptions', $value) && !empty($value['navigationOptions']) ? $value['navigationOptions'] : [];
                    $replaceEntireNavigation = array_key_exists('mergeType', $value) && $value['mergeType'] == AppConstants::MERGE_TYPES['replace'] ? true : false;
                }
            }
        }
        $data = $this->export($modifiedOptions, $tokenType);
        if (!$data) {
            return $data;
        }
        if ($navigationOptions) {
            foreach ($navigationOptions as $navigationOption) {
                $navigationTitleMergeTypeMap[$navigationOption['title']] = $navigationOption['mergeType'];
            }
        }
        foreach ($data as $item) {
            $uuidParts = explode('-', $item['uuid']);
            if ($uuidParts[0] === 'modules') {
                if ($uuidParts[2] === 'settings') {
                    $typeName = 'settings.json';
                } else {
                    $typeName = $uuidParts[2] . '-layout.json';
                }
                $filename = $tmpDirectory . 'modules/' . $this->cleanFileName($uuidParts[1]) . '/' . $typeName;
                $metaData[$uuidParts[1]][] = $uuidParts[2];
            } else {
                if ($item['uuid'] == 'app') {
                    $navigationItemsToReplace = [];
                    $item['uuid'] = 'navigation';
                    $metaData[$item['uuid']] = [];
                    $appendList = [];
                    $navigation = [];
                    foreach ($item['config']['navigation'] as $value) {
                        $navigationTitle = $value['title'];
                        if (in_array($navigationTitle, $appendNavigation)) {
                            $value = $this->sortArrayByKey($value); // First and second level sorting
                            foreach ($value as $key => $arrayElement) {
                                if (is_array($arrayElement) && !empty($arrayElement)) {
                                    $value[$key] = $this->sortArrayByKey($value[$key]); // Third level sorting
                                }
                                foreach ($arrayElement as $arrayKey => $arrayValue) {
                                    if (is_array($arrayValue) && !empty($arrayValue)) {
                                        $arrayElement[$arrayKey] = $this->sortArrayByKey($arrayElement[$arrayKey]); // Fourth level sorting
                                    }
                                }
                                $value[$key] = $arrayElement;
                            }
                            $appendList[] = $value;
                            array_push($navigation, $navigationTitle);
                            if ($navigationTitleMergeTypeMap && array_key_exists($navigationTitle, $navigationTitleMergeTypeMap) && $navigationTitleMergeTypeMap[$navigationTitle] == AppConstants::MERGE_TYPES['replace']) {
                                array_push($navigationItemsToReplace, $navigationTitle);
                            }
                        }
                    }
                    $metaData[$item['uuid']] = $navigation;
                    $item['config']['navigation'] =  $appendList;
                }
                $filename = $tmpDirectory . 'views/' . $item['uuid'] . '.json';
                $navigationItemReplaceOptionsFile = $tmpDirectory . 'views/' . AppConstants::NAVIGATION_OPTION_FOLDER . '/' . AppConstants::NAVIGATION_ITEM_REPLACE_OPTIONS_JSON;
                $navigationItemReplaceOptions['replaceNavigation'] = $replaceEntireNavigation ? true : (count($navigationItemsToReplace) > 0 ? true : false);
                $navigationItemReplaceOptions['navigationItemsToReplace'] = $navigationItemsToReplace;
                $this->writeToFile($navigationItemReplaceOptionsFile, $navigationItemReplaceOptions);
            }
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        if ($type == 'viewTemplates') {
            $viewMetaData = [];
            foreach ($metaData as $key => $value) {
                $mmd = $this->getMetadataByType($key);
                $viewMetaData[$key] = ['name' => $mmd->getDescriptions()['plural'], 'apiName' => $key, 'views' => $value];
            }
            $metaData = $viewMetaData;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        // Module views
        if ($key == 'viewTemplates') {
            $folderPath = $tmpDirectory . 'modules/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $folder) {
                $viewPaths = [
                    $folderPath . $folder . '/detail-layout.json',
                    $folderPath . $folder . '/list-layout.json',
                    $folderPath . $folder . '/form-layout.json',
                    $folderPath . $folder . '/settings.json'
                ];
                foreach ($viewPaths as $viewPath) {
                    if (file_exists($viewPath)) {
                        $fileData = file_get_contents($viewPath);
                        $jsonData = json_decode($fileData, true);
                        $data[] = $jsonData;
                    }
                }
            }
        } else {
            // Other views
            $folderPath = $tmpDirectory . $key . '/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $file) {
                $viewPath = $folderPath . $file;
                if (file_exists($viewPath)) {
                    if (is_dir($viewPath) && $file == AppConstants::NAVIGATION_OPTION_FOLDER) {
                        $viewPath = $viewPath . '/' . AppConstants::NAVIGATION_ITEM_REPLACE_OPTIONS_JSON;
                        $file = AppConstants::NAVIGATION_ITEM_REPLACE_OPTIONS_JSON;
                    }
                    $fileData = file_get_contents($viewPath);
                    $jsonData = json_decode($fileData, true);
                    if (isset($jsonData['uuid']) && $jsonData['uuid'] == 'navigation') {
                        $jsonData['uuid'] = 'app';
                    }
                    if ($file == AppConstants::NAVIGATION_ITEM_REPLACE_OPTIONS_JSON) {
                        $data['navigationItemReplaceOptions'] = $jsonData;
                    } else {
                        $data[] = $jsonData;
                    }
                }
            }
        }

        return $data;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        $navigationOptions = [];
        $navigationItemReplaceOptions = [];
        $fileExtractedPath = $this->tmpFilePath;
        $navigationOptionsJsonPath = $fileExtractedPath . 'views/' . AppConstants::NAVIGATION_OPTION_FOLDER . '/' . AppConstants::NAVIGATION_ITEM_REPLACE_OPTIONS_JSON;
        if (file_exists($navigationOptionsJsonPath)) {
            $jsonData = file_get_contents($navigationOptionsJsonPath);
            $navigationItemReplaceOptions = json_decode($jsonData, true);
        }
        foreach ($options as $value) {
            if ($value['uuid'] == 'app') {
                $value['appendNavigation'] = [];
                $navigationItemsToReplace = $navigationItemReplaceOptions && array_key_exists('navigationItemsToReplace', $navigationItemReplaceOptions) ? $navigationItemReplaceOptions['navigationItemsToReplace'] : [];
                $navMergeType = count($navigationItemsToReplace) == 0 ? AppConstants::MERGE_TYPES['merge'] : (count($value['appendNavigation']) > count($navigationItemsToReplace) ? "" : AppConstants::MERGE_TYPES['replace']);

                foreach ($value['config']['navigation'] as $item) {
                    $value['appendNavigation'][] = $item['title'];
                    $mergeType = in_array($item['title'], $navigationItemsToReplace) ? AppConstants::MERGE_TYPES['replace'] : AppConstants::MERGE_TYPES['merge'];
                    $navigationOptions[] = ['title' => $item['title'], 'mergeType' => $mergeType];
                }
                array_push($exportOption, ['value' => $value['uuid'], 'mergeType' => $navMergeType, 'appendNavigation' => $value['appendNavigation'], 'navigationOptions' => $navigationOptions]);
            } else {
                array_push($exportOption, $value['uuid']);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $item) {
            if (is_array($item)) {
                if ($item["value"] == "app") {
                    $content["navigation"] = $item["appendNavigation"];
                }
            } else {
                $moduleParts = explode('-', $item);
                if (!array_key_exists($moduleParts[1], $content)) {
                    $module = $this->getMetadataByType($moduleParts[1]);
                    $content[$moduleParts[1]] = [
                        "name" => $module->getDescriptions()['plural'],
                        "apiName" => $moduleParts[1],
                        "views" => []
                    ];
                }
                array_push($content[$moduleParts[1]]["views"], $moduleParts[2]);
            }
        }

        return $content;
    }

    public function getMetadataByType($type)
    {
        $repository = $this->entityManager->getRepository(ModelMetadata::class);
        return $repository->findOneBy(['type' => $type]);
    }

    protected function mergeNAppendSubNav($oldSubNavItems, $newSubNavItems)
    {
        $updatedSubNavItems = $oldSubNavItems;
        foreach ($newSubNavItems as $newItem) {
            $itemExists = false;
            foreach ($oldSubNavItems as $key => $oldItem) {
                if ($oldItem['title'] == $newItem['title']) {
                    $itemExists = true;
                    $updatedSubNavItems[$key] = $newItem;
                    break;
                }
            }
            if (!$itemExists) {
                array_push($updatedSubNavItems, $newItem);
            }
        }

        return $updatedSubNavItems;
    }

    protected function mergeNAppendNav($newNav, $oldNav)
    {
        $updatedNav = $oldNav;
        foreach ($newNav as $key => $newNavItem) {
            $include = (array_key_exists("include", $newNavItem) && $newNavItem['include'] == true) || (array_key_exists("include", $newNavItem) && $newNavItem['include'] == "true") || (array_key_exists('mergeType', $newNavItem) && $newNavItem['mergeType'] != AppConstants::MERGE_TYPES['skip']);
            if ($include) {
                $itemExists = false;
                unset($newNavItem['include']);
                unset($newNavItem['exists']);
                foreach ($oldNav['config']['navigation'] as $key => $oldNavItem) {
                    $updatedNavItems = [];
                    if ($oldNavItem['title'] == $newNavItem['title']) {
                        $itemExists = true;
                        $oldNavItems = $oldNavItem['items'] ?? [];
                        $newNavItems = $newNavItem['items'] ?? [];

                        $updatedNav['config']['navigation'][$key] = $newNavItem; //merge values like icon from the new nav
                        // $updatedNav['config']['navigation'][$key]['items'] =  $oldNavItems; //keep the items from old nav, as we would be matching each sub item in seoerate for loop
                        if (empty($oldNavItems) && !empty($newNavItems)) {
                            $updatedNavItems = $newNavItems;
                        }

                        if (!empty($oldNavItems) && !empty($newNavItems)) {
                            $updatedNavItems = $this->mergeNAppendSubNav($oldNavItems, $newNavItems);
                        }
                        if (!empty($updatedNavItems)) {
                            $updatedNav['config']['navigation'][$key]['items'] = $updatedNavItems; //?? incase both old and new item is empty
                        }
                    }
                }
                if (!$itemExists) {
                    array_push($updatedNav['config']['navigation'], $newNavItem);
                }
            }
        }
        return $updatedNav;
    }

    protected function replaceNavItems($newNav, $oldNav)
    {
        $updatedNav = $oldNav;
        # Replace entire navigation
        if ($newNav['replaceEntireNavigation']) {
            return $newNav;
        }
        # Handle replace and append of navigation items as per individual options
        foreach ($newNav['config']['navigation'] as $newNavItemkey => $newNavItem) {
            if ((array_key_exists("include", $newNavItem) && $newNavItem['include'] == true) || (array_key_exists('mergeType', $newNavItem) && $newNavItem['mergeType'] != AppConstants::MERGE_TYPES['skip'])) {
                $itemExists = false;
                foreach ($oldNav['config']['navigation'] as $oldNavItemkey => $oldNavItem) {
                    if ($newNavItem['title'] == $oldNavItem['title']) {
                        $itemExists = true;
                        if (array_key_exists('mergeType', $newNavItem) && $newNavItem['mergeType'] == AppConstants::MERGE_TYPES['replace']) {
                            $updatedNav['config']['navigation'][$oldNavItemkey] = $newNavItem;
                        } else {
                            $oldNavItems = $oldNavItem['items'] ?? [];
                            $newNavItems = $newNavItem['items'] ?? [];
                            $updatedNavItems = $this->mergeNAppendSubNav($oldNavItems, $newNavItems);

                            if (!empty($updatedNavItems)) {
                                $updatedNav['config']['navigation'][$oldNavItemkey]['items'] = $updatedNavItems; //?? incase both old and new item is empty
                            }
                        }
                    }
                }
                if (!$itemExists) {
                    array_push($updatedNav['config']['navigation'], $newNavItem);
                }
            }
        }
        return $updatedNav;
    }

    // Function to check if 'fortiTIP' exists in 'apiName'
    protected function containsFortiTIP($array) {
        foreach ($array as $element) {
            if (isset($element['apiName']) && $element['apiName'] === 'fortiTIP') {
                return true; // Return true if 'fortiTIP' is found
            }
        }
        return false; // Return false otherwise
    }
}
