<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use App\Entity\Core\ConfigProp;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class SKUService
{
    protected $dasClient;
    protected $dasUri;
    private $headers = array("Content-type" => "application/json", "X-ADMIN-USER" => true);

    public function __construct(
        private LoggerInterface $logger,
        protected ContainerInterface $containers,
        protected TagAwareCacheInterface $cache,
        protected EntityManagerInterface $entityManager
    ) {
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->dasUri = $this->containers->getParameter('das.uri');
    }

    public function getLicenseDetailsFromDas()
    {
        $timEnabled = false;
        $isDistributedInstance = false;

        $instanceRole = AppConstants::MASTER_ROLE_NAME;
        $supportedIngestionLimit = AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
        $supportedQueryLimit = AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;

        $productName = $this->getProductNameFromCyopsRelease();
        $productType = AppConstants::PRODUCT_TYPE[$productName];
        $licenseSubType = AppConstants::LICENSE_SUB_TYPE['paid'];

        $dasRequests = new GuzzleRequest('get', $this->dasUri . '/license/?param=license_details&node_id=SELF', $this->headers);
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode < 300) {
                $isDistributedInstance = $jsonDecodeResponse['details']['is_distributed'] ?? false;
                $instanceRole = $isDistributedInstance ? $jsonDecodeResponse['details']['role'][0] : null;
                $nodes = $jsonDecodeResponse['nodes'] ?? [];
                $serialNo = "";
                $subType = "";
                // only current node details will be present in jsonDecodeResponse because of 'node_id=SELF' param in url
                foreach ($nodes as $node) {
                    $serialNo = is_array($node['details']) ? $node['details']['serial_no'] ?? "" : "";
                    $subType = is_array($node['details']) ? $node['details']['sub_type'] ?? "" : "";
                }

                // Contracts key is always expected
                $timEnabled = $jsonDecodeResponse['details']['contracts']['has_tim_subscription'] ?? false;
                $supportedIngestionLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_ingestion_limit'] ?? AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
                $supportedQueryLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_query_limit'] ?? AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;
                $productType = $serialNo && str_starts_with($serialNo, AppConstants::FORTISOAR) ? AppConstants::PRODUCT_TYPE['fortisoar'] : AppConstants::PRODUCT_TYPE['fortitip'];
                $licenseSubType = $subType && $subType == AppConstants::LICENSE_SUB_TYPE_TRIAL ? AppConstants::LICENSE_SUB_TYPE['trial'] : AppConstants::LICENSE_SUB_TYPE['paid'];

                $cacheData = [
                    CacheConstants::IS_DISTRIBUTED_INSTANCE => [
                        'cacheValue' => $isDistributedInstance,
                        'cacheTag' => CacheConstants::MSSP_TYPE_ROLE
                    ],
                    CacheConstants::INSTANCE_ROLE => [
                        'cacheValue' => $instanceRole,
                        'cacheTag' => CacheConstants::MSSP_TYPE_ROLE
                    ],
                    CacheConstants::IS_TIM_ENABLED => [
                        'cacheValue' => $timEnabled,
                        'cacheTag' => CacheConstants::IS_TIM_ENABLED
                    ],
                    CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT => [
                        'cacheValue' => $supportedIngestionLimit,
                        'cacheTag' => CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT
                    ],
                    CacheConstants::TIM_SUPPORTED_QUERY_LIMIT => [
                        'cacheValue' => $supportedQueryLimit,
                        'cacheTag' => CacheConstants::TIM_SUPPORTED_QUERY_LIMIT
                    ],
                    CacheConstants::PRODUCT_TYPE => [
                        'cacheValue' => $productType,
                        'cacheTag' => CacheConstants::PRODUCT_TYPE
                    ],
                    CacheConstants::SUB_TYPE => [
                        'cacheValue' => $licenseSubType,
                        'cacheTag' => CacheConstants::SUB_TYPE
                    ]
                ];
                $this->setCache($cacheData);
                return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit, $productType, $licenseSubType];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->info($error->getMessage());
        }
        return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit, $productType, $licenseSubType];
    }

    public function getHAClusterNodesFromDas()
    {
        $nodeList = array();
        $retryCount = 5;
        while ($retryCount > 0) {
            try {
                $cachedNodes = $this->cache->getItem("haNodes");
                if (!$cachedNodes->isHit()) {
                    $dasRequests = new GuzzleRequest('get', $this->dasUri . '/cluster/', ["Content-type" => "application/json"]);
                    $dasResponse = $this->dasClient->send($dasRequests);
                    $resCode = $dasResponse->getStatusCode();
                    if ($resCode != 200) {
                        error_log('Failed to fetch ha nodes from das');
                    }
                    $dasResponseData = $dasResponse->getBody()->getContents();
                    $jsonDecodeResponse = json_decode($dasResponseData, true);
                    $fortisoarNodes = $jsonDecodeResponse["nodes"];
                    foreach ($fortisoarNodes as $node) {
                        if ($node["status"] == "active") {
                            $nodeList[] = $node["nodeName"];
                        }
                    }
                    if (!empty($nodeList)) {
                        $cachedNodes->set($nodeList);
                        $cachedNodes->expiresAfter(7200);
                        $cachedNodes->tag(CacheConstants::HA_NODES);
                        $this->cache->save($cachedNodes);
                    } else {
                        throw new \Exception("HA cluster node list cannot be empty");
                    }
                }
                $nodeList = $cachedNodes->get();
                break;
            } catch (\Exception $error) {
                error_log('cyops-auth service request failed');
                error_log($error->getMessage());
                sleep(2);
                $retryCount -= 1;
            }
        }

        return $nodeList;
    }

    public function setCache($cacheInfo)
    {
        foreach ($cacheInfo as $cacheKey => $cacheData) {
            $cacheValue = $cacheData['cacheValue'];
            $cacheTag = $cacheData['cacheTag'];
            $item = $this->cache->getItem($cacheKey);
            $item->set($cacheValue);
            $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
            $item->tag($cacheTag);
            $this->cache->save($item);
        }
    }

    public function getProductFeatureControlData($repoLink, $productTypeKey)
    {
        try {
            $featureCachedObject = $this->getCacheRepoData($productTypeKey);
            if ($featureCachedObject && $this->isCacheValid($featureCachedObject->getModifiedAt())) {
                return json_decode($this->getCacheRepoData($productTypeKey)->getValue(), true);
            }
            $client = new \GuzzleHttp\Client();
            $featureControlJsonObject = $client->request('GET', $repoLink, ['verify' => false]);
            $productFeatureControlConfig = json_decode($featureControlJsonObject->getBody()->getContents(), true);
            $this->cacheRepoServiceResponse($productFeatureControlConfig, $featureCachedObject, $productTypeKey);
            return $productFeatureControlConfig;
        } catch (\Exception $ex) {
            $this->logger->error($ex->getMessage());
            $featureCachedObject = $this->getCacheRepoData($productTypeKey);
            if (is_null($featureCachedObject)) {
                $this->logger->error(ErrorConstants::FSR_CH_0000024);
                throw new \Exception(ErrorConstants::FSR_CH_0000024);
            }
            return json_decode($featureCachedObject->getValue(), true);
        }
    }

    private function isCacheValid($cachedTime)
    {
        $currentTime = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $cachedTime->setTimezone(new \DateTimeZone('UTC'));
        // Computing the difference in seconds
        $secondsDifference = $currentTime->getTimestamp() - $cachedTime->getTimestamp();
        return $secondsDifference < AppConstants::REPO_DATA_CACHE_EXPIRY_SECONDS;
    }

    private function getCacheRepoData($productTypeKey)
    {
        $cachedProp = $this->entityManager->getRepository(ConfigProp::class)
            ->findOneBy(['key' => $productTypeKey]);
        return $cachedProp;
    }

    private function cacheRepoServiceResponse($productFeatureControlData, $existingProp, $productTypeKey)
    {
        $now = new \DateTime('NOW', new \DateTimeZone('UTC'));
        if ($existingProp === null) {
            // If no cache, create new entry
            $persistedProp = new ConfigProp();
            $persistedProp->setKey($productTypeKey);
            $persistedProp->setValue(json_encode($productFeatureControlData));
            $persistedProp->setModifiedAt($now);
            $this->entityManager->persist($persistedProp);
        } else {
            // Update existing cache with new data and timestamp
            $existingProp->setValue(json_encode($productFeatureControlData));
            $existingProp->setModifiedAt($now);
            $this->entityManager->persist($existingProp);
        }
        $this->entityManager->flush();
    }

    public function fetchLicenseBasedFeatureDetails($productFeatureControlJson, $productTypeKey = null, $componentType = null)
    {
        $repoServer = getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server");
        $repoLink = sprintf(AppConstants::FEATURE_CONTROL_URL, $repoServer, $productFeatureControlJson);
        $productFeatureControlData = $this->getProductFeatureControlData($repoLink, $productTypeKey);
        $licenseBasedFeatureData = $productTypeKey && array_key_exists($productTypeKey, $productFeatureControlData) ?
            $productFeatureControlData[$productTypeKey] : $productFeatureControlData;
        if ($componentType && array_key_exists($componentType, $licenseBasedFeatureData)) {
            $licenseBasedFeatureData = $licenseBasedFeatureData[$componentType];
        }
        return $licenseBasedFeatureData;
    }

    public function getProductNameFromCyopsRelease()
    {
        $productName = null;
        $cyopsReleaseFileData = file_get_contents(AppConstants::CYOPS_RELEASE_PATH);
        if ($cyopsReleaseFileData !== false) {
            if (preg_match(AppConstants::PRODUCT_NAME_REGEX, $cyopsReleaseFileData, $matches)) {
                $productName = trim($matches[1]);
            }
        }
        return $productName;
    }
}
